/*  -------------------------------------------------------------------  */
/*  NOTE: The _SP_base macro should be defined to indicate the offset    */
/*        of the sparse storage index vectors  (0 or 1)                  */
/*  -------------------------------------------------------------------  */

/*
//                          Sparse BLAS in C
//                             Version 1.1
//       [ Optimization of v. 1.0 for varying vector offsets, plus  ]
//       [   functionality for matvecs with symmetric matrices      ]
//
//    Developed in conjunction with the SparseLib++ Sparse Matrix Library
//
//             National Institute of Standards and Technology, MD.
//                         University of Notre Dame, IN.
//              Authors:  R. Pozo, K. Remington, A. Lumsdaine
//                     (C) 1995 All Rights Reserved
//
//                                NOTICE
//
// Permission to use, copy, modify, and distribute this software and
// its documentation for any purpose and without fee is hereby granted
// provided that the above copyright notice appear in all copies and
// that both the copyright notice and this permission notice appear in
// supporting documentation.
//
// Neither the Authors nor the Institutions (National Institute of Standards
// and Technology, University of Notre Dame) make any representations about
// the suitability of this software for any purpose. This software is
// provided "as is" without expressed or implied warranty.
//
// SparseLib++ was funded in part by the U.S. Department of Energy, the
// National Science Foundation, the University of Notre Dame, and the State
// of Tennessee.
*/

#include <stdlib.h>
#include <stdio.h>
#include "spblas.h"

void CoordMatVec_TYPE(int m, int n, int k, const TYPE alpha,
                 const TYPE *val, const int *indx, const int *jndx,
                 const int nnz, const TYPE *b, int ldb, const TYPE beta, 
                 TYPE *c, int ldc, TYPE *work, const int lwork, const int ind_base)
{
  /*
  // C <- alpha*A*B + beta*C
  // C is m x n (Dense)
  // A is m x k (Sparse; contained in val[] with index arrays indx[] and jndx[])
  // B is k x n (Dense)
  // alpha and beta are scalars
  */

  int i, j, l;

  /*
  // References to k and lwork to avoid compiler warnings
  // (needed in interface, but not currently used)
  */

  if (k&&lwork) { };


  if (alpha == 0.0) {
     ScaleArray_TYPE(m, n, c, ldc, beta);
     return;
  }


  if (alpha == 1.0) {
    if (n == 1) {
/*    Just one matrix * vector, with no scaling:    */
      if (beta != 0 ) for (i=0;i<m;i++) work[i] = beta*c[i];
      c=c-ind_base;
      b=b-ind_base;
      for (j = 0; j < nnz; j++)
	c[indx[j]] += b[jndx[j]] * val[j];
      if (beta != 0 ) {
         c=c+ind_base;
         for (i=0;i<m;i++) c[i] += work[i];
      }
    } else  {
/*    Multiple matrix * vector, with no scaling:    */
      for (i = 0; i < n; i++){
        if (beta != 0 ) 
            for (l=0;l<m;l++) 
               work[l] = beta*c[ldc*i+l];
        c=c-ind_base;
        b=b-ind_base;
	for (j = 0; j < nnz; j++)
	  c[ldc*i+indx[j]] += b[ldb*i+indx[j]] * val[j];
        c=c+ind_base;
        b=b+ind_base;
        if (beta != 0 ) {
            for (l=0;l<m;l++) 
               c[ldc*i+l] += work[l];
        }
      }
    }
  } else {
    if (n == 1) {
/*    Just one matrix * vector, with scaling:    */
      if (beta != 0 ) for (i=0;i<m;i++) work[i] = beta*c[i];
      c=c-ind_base;
      b=b-ind_base;
      for (j = 0; j < nnz; j++)
	c[indx[j]] += alpha * b[jndx[j]] * val[j];
      if (beta != 0 ) {
        c=c+ind_base;
        for (i=0;i<m;i++) c[i] += work[i];
      }
    } else {
/*    Multiple matrix * vector, with scaling:    */
      for (i = 0; i < n; i++){
        if (beta != 0 ) 
            for (l=0;l<m;l++) 
               work[l] = beta*c[ldc*i+l];
        c=c-ind_base;
        b=b-ind_base;
	for (j = 0; j < nnz; j++)
	    c[ldc*i+indx[j]] += alpha * b[ldb*i+indx[j]] * val[j];
        c=c+ind_base;
        b=b+ind_base;
        if (beta != 0 ) {
            for (l=0;l<m;l++) 
               c[ldc*i+l] += work[l];
        }
      }
    }
  }
}

void COSymmMatVec_TYPE(int m, int n, int k, const TYPE alpha,
                 const TYPE *val, const int *indx, const int *jndx,
                 const int nnz, const TYPE *b, int ldb, const TYPE beta, 
                 TYPE *c, int ldc, TYPE *work, const int lwork, const int ind_base)
{
  /*
  // C <- alpha*A*B + beta*C
  // C is m x n (Dense)
  // A is m x k (Symmetric Sparse; contained in val[] with index arrays indx[] and jndx[])
  // B is k x n (Dense)
  // alpha and beta are scalars
  */

  int i, j, l;
  int  rowi, coli;

  /*
  // References to k and lwork to avoid compiler warnings
  // (needed in interface, but not currently used)
  */

  if (k&&lwork) { };

  if (alpha == 0.0) {
     ScaleArray_TYPE(m, n, c, ldc, beta);
     return;
  }

  if (alpha == 1.0) {
    if (n == 1) {
/*    Just one matrix * vector, with no scaling:    */
      if (beta != 0 ) for (i=0;i<m;i++) work[i] = beta*c[i];
      c=c-ind_base;
      b=b-ind_base;
      for (j = 0; j < nnz; j++) {
           rowi =  indx[j];
           coli =  jndx[j];
           if ( rowi ==  coli )
                c[rowi]  += val[j]*b[coli];
           else {
                c[rowi]  += val[j]*b[coli];
                c[coli]  += val[j]*b[rowi];
           }
      }
      c=c+ind_base;
      if (beta != 0 ) {
         for (i=0;i<m;i++) c[i] += work[i];
      }
    } else  {
/*    Multiple matrix * vector, with no scaling:    */
      for (i = 0; i < n; i++){
        if (beta != 0 ) 
            for (l=0;l<m;l++) 
               work[l] = beta*c[ldc*i+l];
        c=c-ind_base;
        b=b-ind_base;
        for (j = 0; j < nnz; j++) {
           rowi =  indx[j];
           coli =  jndx[j];
           if ( rowi ==  coli )
	        c[ldc*i+rowi] += b[ldb*i+coli] * val[j];
           else {
	        c[ldc*i+rowi] += b[ldb*i+coli] * val[j];
	        c[ldc*i+coli] += b[ldb*i+rowi] * val[j];
           }
        }
        c=c+ind_base;
        b=b+ind_base;
        if (beta != 0 )  {
            for (l=0;l<m;l++) 
               c[ldc*i+l] += work[l];
        }
      }
    }
  } else {
    if (n == 1) {
/*    Just one matrix * vector, with scaling:    */
      if (beta != 0 ) for (i=0;i<m;i++) work[i] = beta*c[i];
      c=c-ind_base;
      b=b-ind_base;
      for (j = 0; j < nnz; j++) {
           rowi =  indx[j];
           coli =  jndx[j];
           if ( rowi ==  coli )
                c[rowi]  += alpha*val[j]*b[coli];
           else {
                c[rowi]  += alpha*val[j]*b[coli];
                c[coli]  += alpha*val[j]*b[rowi];
           }
      }
      c=c+ind_base;
      if (beta != 0 ) {
         for (i=0;i<m;i++) c[i] += work[i];
      }
    } else {
/*    Multiple matrix * vector, with scaling:    */
      for (i = 0; i < n; i++){
        if (beta != 0 ) 
            for (l=0;l<m;l++) 
               work[l] = beta*c[ldc*i+l];
        c=c-ind_base;
        b=b-ind_base;
        for (j = 0; j < nnz; j++) {
           rowi =  indx[j];
           coli =  jndx[j];
           if ( rowi ==  coli )
	        c[ldc*i+rowi] += 
	               alpha*b[ldb*i+coli] * val[j];
           else {
	        c[ldc*i+rowi] += alpha*b[ldb*i+coli] * val[j];
	        c[ldc*i+coli] += alpha*b[ldb*i+rowi] * val[j];
           }
        }

	for (j = 0; j < nnz; j++)
	   c[ldc*i+indx[j]] += alpha * b[ldb*i+indx[j]] * val[j];
        c=c+ind_base;
        b=b+ind_base;
        if (beta != 0 ) {
            for (l=0;l<m;l++) 
               c[ldc*i+l] += work[l];
         }
      }
    }
  }
}

/*-----------------------------------------------------------------------------*/
/*-----------------------------------------------------------------------------*/

/* F77 interface routine: */
void  _coomm(const int transa, const int m, const int n, const int k,
             const TYPE alpha, const int descra[], const TYPE *val,
             const int *indx, const int *jndx, const int nnz,
             const TYPE *b, const int ldb,
             const TYPE beta, TYPE *c, const int ldc,
             TYPE *work, const int lwork)
{
/*-----------------------------------------------------------------------------*/
/*-----------------------------------------------------------------------------*/
/* F77 interface:
 * _coomm -- coordinate format matrix-matrix multiply
 *
 * C <- alpha A B + beta C
 *
 * Arguments:
 *
 * int &transa	Indicates how to operate with the sparse matrix
 *		0 : operate with matrix
 *		1 : operate with transpose matrix
 *		2 : operate with conjugate transpose matrix
 *
 * int &m	Number of rows in matrix c
 *
 * int &n	Number of columns in matrix c
 *
 * int &k	Number of rows in matrix b
 *
 * TYPE		 &alpha Scalar parameter
 *
 * TYPE		 &beta Scalar parameter
 *
 * int descra[]	Descriptor argument.  Nine element integer array
 *		descra[0] matrix structure
 *			0 : general
 *			1 : symmetric
 *			2 : Hermitian
 *			3 : Triangular
 *			4 : Anti-Symmetric
 *			5 : Diagonal
 *		descra[1] upper/lower triangular indicator
 *			1 : lower
 *			2 : upper
 *		descra[2] main diagonal type
 *			0 : non-unit
 *			1 : unit
 *		descra[4] repeated indices?
 *			0 : unknown
 *			1 : no repeated indices
 *
 *
 * TYPE		*val	scalar array of length nnz containing matrix entries
 *
 * int		*indx	integer array of length nnz containing row indices
 *
 * int		*jndx	integer array of length nnz containing column indices
 *
 * TYPE		*b	rectangular array with first dimension ldb
 *
 * TYPE		*c	rectangular array with first dimension ldc
 *
 * TYPE		*work	scratch array of length lwork.  lwork should be at least
 *		max(m,n)
 *
 */
/*-----------------------------------------------------------------------------*/
/*-----------------------------------------------------------------------------*/
  if (descra[0] == 2 || descra[0] == 4) {
    fprintf(stderr,"Matrix type descra[0] == %d ",descra[0]);
    fprintf(stderr,"not currently supported in _coomm.  Supported values: 0,1,3,5\n");
    exit(1);
  }
  if (alpha == 0.0) {
        ScaleArray_TYPE(m, n, c, ldc, beta);
        return;
  }

  if (descra[0] == 1) {
    if (descra[1] == 1 )
    /* Symmetric coordinate matrix stored in lower triangular part */
       COSymmMatVec_TYPE(m, n, k, alpha, val, jndx, indx, nnz, &b[0], ldb, 
                                  beta, &c[0], ldc, &work[0], lwork, _SP_base);
    else
    /* Symmetric coordinate matrix stored in upper triangular part */
       COSymmMatVec_TYPE(m, n, k, alpha, val, indx, jndx, nnz, &b[0], ldb, 
                                  beta, &c[0], ldc, &work[0], lwork, _SP_base);
  } else {
    /* If transpose is desired, interchange row and column indices: */
    if (transa == 1 || transa == 2) {
    CoordMatVec_TYPE(m, n, k, alpha, val, jndx, indx, nnz, &b[0], ldb, 
                                  beta, &c[0], ldc, &work[0], lwork, _SP_base);
    } else {
    CoordMatVec_TYPE(m, n, k, alpha, val, indx, jndx, nnz, &b[0], ldb, 
                                  beta, &c[0], ldc, &work[0], lwork, _SP_base);
    }
  }
}

