/*  -------------------------------------------------------------------  */
/*  NOTE: The _SP_base macro should be defined to indciate the offset    */
/*        of the sparse storage index vectors  (0 or 1)                  */
/*  -------------------------------------------------------------------  */

/*
//                          Sparse BLAS in C
//                             Version 1.1
//       [ Optimization of v. 1.0 for varying vector offsets, plus  ]
//       [   functionality for matvecs with symmetric matrices      ]
//
//    Developed in conjunction with the SparseLib++ Sparse Matrix Library
//
//             National Institute of Standards and Technology, MD.
//                         University of Notre Dame, IN.
//              Authors:  R. Pozo, K. Remington, A. Lumsdaine
//                     (C) 1995 All Rights Reserved
//
//                                NOTICE
//
// Permission to use, copy, modify, and distribute this software and
// its documentation for any purpose and without fee is hereby granted
// provided that the above copyright notice appear in all copies and
// that both the copyright notice and this permission notice appear in
// supporting documentation.
//
// Neither the Authors nor the Institutions (National Institute of Standards
// and Technology, University of Notre Dame) make any representations about
// the suitability of this software for any purpose. This software is
// provided "as is" without expressed or implied warranty.
//
// SparseLib++ was funded in part by the U.S. Department of Energy, the
// National Science Foundation, the University of Notre Dame, and the State
// of Tennessee.
*/

#include <stdlib.h>
#include <stdio.h>
#include "spblas.h"

/* F77 interface routine: */
void _coosm (const int transa, const int m, const int n,
   const int unitd, const TYPE *dv, const TYPE alpha,
   const int descra[], const TYPE *val,
   const int *indx, const int *jndx, const int nnz, const TYPE *b, int ldb,
   const TYPE beta, TYPE *c, const int ldc,
   TYPE *work, const int lwork)
{
  /*
  // C <-  alpha*D*A^{-1}*B + beta*C
  // C <-  alpha*A^{-1}*D*B + beta*C
  // C is m x n
  // A is m x m  (Sparse; contained in val[] with index arrays indx[] and jndx[])
  // B is m x n
  // D is m x m (Diagonal; values contained in dv[]0
  // unitd = 1   D = I
  // unitd = 2   left (row scaling)
  // unitd = 3   right (column scaling)
  // alpha is a scalar
  */
  int i,j;
  int *tally; 
  int *colptr; 
  int *rowind;
  TYPE *cval;

  int ind_base = _SP_base;

  if (descra[0] != 3) {
    fprintf(stderr,"Must have triangular matrix for _coosm. (descra[0] == 3)\n");
    exit(1);
  }

/* First, convert storage vectors to CompCol format: */
/* (adapted from CompCol_Mat_TYPE(Coord_Mat_TYPE) constructor) */
        tally  = (int *)malloc((m+1)*sizeof(int));
        for (i=0;i<m+1;i++) tally[i] = 0;
        colptr = (int *)malloc((m+1)*sizeof(int));
        rowind = (int *)malloc((nnz)*sizeof(int));
        cval = (TYPE *)malloc((nnz)*sizeof(TYPE));
/*      The following pointers will be incremented/decremented to create         */
/*      the appropriate vector offset if ind_base != 0 (e.g. 1-based vectors)    */
/*      This elminates the need to adjust element by element with ind_base.      */
        tally=tally-ind_base;
        rowind=rowind-ind_base;
        cval=cval-ind_base;
/*      First pass through nonzeros.  Tally entries in each column.         */
/*      And calculate colptr array. */
        for (i=0;i<nnz;i++) tally[jndx[i]]++;
        colptr[0] = ind_base;
        tally=tally+ind_base;
        for (j=0;j<m;j++) colptr[j+1] = colptr[j]+tally[j];
/*      Make copy of colptr for use in second pass. */
        for (j=0;j<m;j++) tally[j] = colptr[j];
        tally=tally-ind_base;
/*      Second pass through nonzeros.   Fill in index and value entries. */
        for (i=0;i<nnz;i++)
        {
           cval[tally[jndx[i]]] = val[i];
           rowind[tally[jndx[i]]] = indx[i];
           tally[jndx[i]]++;
        }
        tally=tally+ind_base;
        rowind=rowind+ind_base;
        cval=cval+ind_base;

/* Now, call the appropriate routines using the newly created CompCol storage vectors: */

  if (transa == 0) {
    if (descra[1] == 1) {
      if (descra[2] == 0)
	CompCol_LD_TYPE(m, n, unitd, dv, alpha, cval, 
                         rowind, colptr, b, ldb, beta, c, ldc,
                         work, lwork, _SP_base);
      else if (descra[2] == 1)
	CompCol_LU_TYPE(m, n, unitd, dv, alpha, cval, 
                         rowind, colptr, b, ldb, beta, c, ldc,
                         work, lwork, _SP_base);
      else {
	fprintf(stderr,"Bad descra[2] in _coosm.\n");
        free(cval); free(colptr); free(rowind); free(tally);
	exit(1);
      }
    } else if (descra[1] == 2) {
      if (descra[2] == 0)
	CompCol_UD_TYPE(m, n, unitd, dv, alpha, cval, 
                         rowind, colptr, b, ldb, beta, c, ldc,
                         work, lwork, _SP_base);
      else if (descra[2] == 1)
	CompCol_UU_TYPE(m, n, unitd, dv, alpha, cval, 
                         rowind, colptr, b, ldb, beta, c, ldc,
                         work, lwork, _SP_base);
      else {
	fprintf(stderr,"Bad descra[2] in _coosm.\n");
        free(cval); free(colptr); free(rowind); free(tally);
	exit(1);
      }
    }
  } else if (transa == 1 || transa == 2) {
    if (descra[1] == 1) {
      if (descra[2] == 0)
	CompRow_UD_TYPE(m, n, unitd, dv, alpha, cval, 
                         rowind, colptr, b, ldb, beta, c, ldc,
                         work, lwork, _SP_base);
      else if (descra[2] == 1)
	CompRow_UU_TYPE(m, n, unitd, dv, alpha, cval, 
                         rowind, colptr, b, ldb, beta, c, ldc,
                         work, lwork, _SP_base);
      else {
	fprintf(stderr,"Bad descra[2] in _coosm.\n");
        free(cval); free(colptr); free(rowind); free(tally);
	exit(1);
      }
    } else if (descra[1] == 2) {
      if (descra[2] == 0)
	CompRow_LD_TYPE(m, n, unitd, dv, alpha, cval, 
                         rowind, colptr, b, ldb, beta, c, ldc,
                         work, lwork, _SP_base);
      else if (descra[2] == 1)
	CompRow_LU_TYPE(m, n, unitd, dv, alpha, cval, 
                         rowind, colptr, b, ldb, beta, c, ldc,
                         work, lwork, _SP_base);
      else {
	fprintf(stderr,"Bad descra[2] in _coosm.\n");
        free(cval); free(colptr); free(rowind); free(tally);
	exit(1);
      }
    } else {
      fprintf(stderr,"Bad descra[1] in _coosm.\n");
      free(cval); free(colptr); free(rowind); free(tally);
      exit(1);
    }
  } else {
    fprintf(stderr,"Bad transa in _coosm.\n");
    free(cval); free(colptr); free(rowind); free(tally);
    exit(1);
  }
  free(cval); free(colptr); free(rowind); free(tally);
}

