/*  -------------------------------------------------------------------  */
/*  NOTE: The _SP_base macro should be defined to indciate the offset    */
/*        of the sparse storage index vectors  (0 or 1)                  */
/*  -------------------------------------------------------------------  */

/*
//                          Sparse BLAS in C
//                             Version 1.1
//       [ Optimization of v. 1.0 for varying vector offsets, plus  ]
//       [   functionality for matvecs with symmetric matrices      ]
//
//    Developed in conjunction with the SparseLib++ Sparse Matrix Library
//
//             National Institute of Standards and Technology, MD.
//                         University of Notre Dame, IN.
//              Authors:  R. Pozo, K. Remington, A. Lumsdaine
//                     (C) 1995 All Rights Reserved
//
//                                NOTICE
//
// Permission to use, copy, modify, and distribute this software and
// its documentation for any purpose and without fee is hereby granted
// provided that the above copyright notice appear in all copies and
// that both the copyright notice and this permission notice appear in
// supporting documentation.
//
// Neither the Authors nor the Institutions (National Institute of Standards
// and Technology, University of Notre Dame) make any representations about
// the suitability of this software for any purpose. This software is
// provided "as is" without expressed or implied warranty.
//
// SparseLib++ was funded in part by the U.S. Department of Energy, the
// National Science Foundation, the University of Notre Dame, and the State
// of Tennessee.
*/

#include <stdlib.h>
#include <stdio.h>
#include "spblas.h"


void CompRowMatVec_TYPE(int m, int n, int kk, const TYPE alpha,
                          const TYPE *val, const int *indx, const int *pntr,
                          const TYPE *b, int ldb, const TYPE beta, TYPE *c, 
                          int ldc, TYPE *work, const int lwork, const int ind_base)
{
  /*
  // C <- alpha*A*B + beta*C
  // C is m x n (Dense)
  // A is m x k (Sparse; contained in val[] with index arrays indx[] and jndx[])
  // B is k x n (Dense)
  // alpha and beta are scalars
  */

  int i, j, k, l;

  /*
  // Reference to k and lwork to avoid compiler warnings
  // (needed in interface, but not currently used)
  */
  if (kk&&lwork) { };

  if (alpha == 0.0) {
     ScaleArray_TYPE(m, n, c, ldc, beta);
     return;
  }

  val=val-ind_base;
  b=b-ind_base;
  indx=indx-ind_base;

  if (alpha == 1.0) {
    if (n == 1)
/*    Just one matrix * vector, with no scaling:        */
      if (beta != 0 ) for (i=0;i<m;i++) work[i] = beta*c[i];
      for (i = 0; i < m; i++) {
        c[i] = 0;
	for (j = pntr[i]; j < pntr[i+1]; j++)
	  c[i] += b[indx[j]] * val[j];
      }
      if (beta != 0 ) for (i=0;i<m;i++) c[i] += work[i];
    else
/*    Multiple matrix * vector, with no scaling:       */
      for (l = 0; l < n; l++) {
        if (beta != 0 ) 
            for (k=0;k<m;k++) 
               work[k] = beta*c[ldc*i+k];
	for (i = 0; i < m; i++) {
          c[ldc*l+i] = 0;
	  for (j = pntr[i]; j < pntr[i+1]; j++)
	    c[ldc*l+i] += b[ldb*l+indx[j]] * val[j];
        }
        if (beta != 0 ) 
            for (k=0;k<m;k++) 
               c[ldc*i+k] += work[k];
      }
  } else {
    if (n == 1)
/*    Just one matrix * vector, with scaling:        */
      if (beta != 0 ) for (i=0;i<m;i++) work[i] = beta*c[i];
      for (i = 0; i < m; i++) {
        c[i] = 0;
	for (j = pntr[i]; j < pntr[i+1]; j++)
	  c[i] += alpha * b[indx[j]] * val[j];
      }
      if (beta != 0 ) for (i=0;i<m;i++) c[i] += work[i];
    else
/*    Multiple matrix * vector, with scaling:        */
      for (l = 0; l < n; l++) {
        if (beta != 0 ) 
            for (k=0;k<m;k++) 
               work[k] = beta*c[ldc*i+k];
	for (i = 0; i < m; i++) {
          c[ldc*l+i] = 0;
	  for (j = pntr[i]; j < pntr[i+1]; j++)
	    c[ldc*l+i] +=
	      alpha * b[ldb*l+indx[j]] * val[j];
        }
        if (beta != 0 ) 
            for (k=0;k<m;k++) 
               c[ldc*i+k] += work[k];
      }
  }
}

void CRSymmMatVec_TYPE(int m, int n, int k, const TYPE alpha,
		   const TYPE *val, const int *indx, const int *pntr,
		   const TYPE *b, int ldb, const TYPE beta, TYPE *c, int ldc,
                   TYPE *work, const int lwork, const int ind_base)
{
  /*
  // C <- alpha*A*B + beta*C
  // C is m x n (Dense)
  // A is m x k (Symmetric Sparse; contained in val[] with index arrays indx[] and jndx[])
  // B is k x n (Dense)
  // alpha and beta are scalars
  */

  int i, j, l;
  int row_ptrlow, row_ptrhi;

  /* Reference to m and lwork to avoid compiler warnings */
  /* (needed in interface, but not currently used)       */
  if (m&&lwork) { };

  if (alpha == 0.0) {
     ScaleArray_TYPE(m, n, c, ldc, beta);
     return;
  }

  val=val-ind_base;
  indx=indx-ind_base;

  if (alpha == 1.0) {
    if (n == 1) {
/*    Just one matrix * vector, with no scaling:   */
      if (beta != 0 ) for (i=0;i<m;i++) work[i] = beta*c[i];
      for (i=0;i<m;i++) c[i] = 0;
      for(j=0;j<k;j++){
          row_ptrlow = pntr[j];
          row_ptrhi = pntr[j+1];
          if ( indx[row_ptrhi-1] == j+ind_base ) {
             c[j] += b[j]*val[row_ptrhi-1];
             row_ptrhi--;
          }
          b=b-ind_base;
          for(i=row_ptrlow;i<row_ptrhi;i++){
             c[j] += b[indx[i]]*val[i];
          }
          c=c-ind_base;
          b=b+ind_base;
          for(i=row_ptrlow;i<row_ptrhi;i++){
             c[indx[i]]+= b[j]*val[i];
          }
          c=c+ind_base;
       }
      if (beta != 0 ) for (i=0;i<m;i++) c[i] += work[i];
    } else {
/*    Multiple matrix * vector, with no scaling:   */
      for (l = 0; l < n; l++){
        if (beta != 0 ) 
            for (i=0;i<m;i++) work[i] = beta*c[ldc*l+i];
        for (i=0;i<m;i++) c[ldc*l+i] = 0;
        for(j=0;j<k;j++){
          row_ptrlow = pntr[j];
          row_ptrhi = pntr[j+1];
          if ( indx[row_ptrhi-1] == j+ind_base ) {
	     c[ldc*l+j] += b[ldb*l+j] * val[row_ptrhi-1];
             row_ptrhi--;
          }
          b=b-ind_base;
          for(i=row_ptrlow;i<row_ptrhi;i++){
	     c[ldc*l+j] += b[ldb*l+indx[i]] * val[i];
          }
          c=c-ind_base;
          b=b+ind_base;
          for(i=row_ptrlow;i<row_ptrhi;i++){
	     c[ldc*l+indx[i]] += b[ldb*l+j] * val[i];
          }
          c=c+ind_base;
        }
        if (beta != 0 ) 
            for (i=0;i<m;i++) c[ldc*l+i] += work[i];
      }
    }
  } else {
    if (n == 1) {
/*    Just one matrix * vector, with scaling:    */
      if (beta != 0 ) for (i=0;i<m;i++) work[i] = beta*c[i];
      for (i=0;i<m;i++) c[i] = 0;
      for(j=0;j<k;j++){
          row_ptrlow = pntr[j];
          row_ptrhi = pntr[j+1];
          if ( indx[row_ptrhi-1] == j+ind_base ) {
             c[j] += alpha*b[j]*val[row_ptrhi-1];
             row_ptrhi--;
          }
          b=b-ind_base;
          for(i=row_ptrlow;i<row_ptrhi;i++){
             c[j] += alpha*b[indx[i]]*val[i];
          }
          c=c-ind_base;
          b=b+ind_base;
          for(i=row_ptrlow;i<row_ptrhi;i++){
             c[indx[i]]+= alpha*b[j]*val[i];
          }
          c=c+ind_base;
       }
      if (beta != 0 ) for (i=0;i<m;i++) c[i] += work[i];
    } else {
/*    Multiple matrix * vector, with scaling:    */
      for (l = 0; l < n; l++) {
        if (beta != 0 ) for (i=0;i<m;k++) work[i] = beta*c[ldc*l+i];
        for (i=0;i<m;i++) c[ldc*l+i] = 0;
        for(j=0;j<k;j++){
          row_ptrlow = pntr[j];
          row_ptrhi = pntr[j+1];
          if ( indx[row_ptrhi-1] == j+ind_base ) {
	     c[ldc*l+j] += alpha*b[ldb*l+j] * val[row_ptrhi-1];
             row_ptrhi--;
          }
          b=b-ind_base;
          for(i=row_ptrlow;i<row_ptrhi;i++){
	     c[ldc*l+j] += alpha*b[ldb*l+indx[i]] * val[i];
          }
          c=c-ind_base;
          b=b+ind_base;
          for(i=row_ptrlow;i<row_ptrhi;i++){
	     c[ldc*l+indx[i]] += alpha*b[ldb*l+j] * val[i];
          }
          c=c+ind_base;
        }
        if (beta != 0 ) for (i=0;i<m;i++) c[ldc*l+i] += work[i];
      }
    }
  }
}
/*-----------------------------------------------------------------------------*/
/*-----------------------------------------------------------------------------*/

/* F77 interface routine: */
void _csrmm(const int transa, const int m, const int n, const int k,
            const TYPE alpha, const int descra[], const TYPE *val,
            const int *indx, const int *pntr, const TYPE *b, int ldb,
            const TYPE beta, TYPE *c, const int ldc,
            TYPE *work, const int lwork)
{
/*-----------------------------------------------------------------------------*/
/*-----------------------------------------------------------------------------*/
/* F77 interface:
 * _csrmm -- comp sparse row matrix-matrix multiply
 *
 * Arguments:
 *
 * int &transa	Indicates how to operate with the sparse matrix
 *		0 : operate with matrix
 *		1 : operate with transpose matrix
 *		2 : operate with conjugate transpose matrix
 *
 * int &m	Number of rows in matrix c
 *
 * int &n	Number of columns in matrix c
 *
 * int &k	Number of rows in matrix b
 *
 * TYPE		&alpha Scalar parameter
 *
 * TYPE		&beta Scalar parameter
 *
 * int descra[]	Descriptor argument.  Nine element integer array
 *		descra[0] matrix structure
 *			0 : general
 *			1 : symmetric
 *			2 : Hermition
 *			3 : Triangular
 *			4 : Anti-Symmetric
 *			5 : Diagonal
 *		descra[1] upper/lower triangular indicator
 *			1 : lower
 *			2 : upper
 *		descra[2] main diagonal type
 *			0 : non-unit
 *			1 : unit
 *
 * TYPE		*val	scalar array of length nnz containing matrix entries
 *
 * int *indx	integer array of length nnz containing column indices
 *
 * int *pntr	integer array of length k+1 such that pntr(j)-pntr(1)
 *		points to location in val of the first element in row j
 *
 * TYPE		*b	rectangular array with first dimension ldb
 *
 * TYPE		*c	rectangular array with first dimension ldc
 *
 * TYPE		*work	scratch array of length lwork.  lwork should be at least
 *		max(m,n)
 *
 */
/*-----------------------------------------------------------------------------*/
/*-----------------------------------------------------------------------------*/
  if (descra[0] == 2 || descra[0] == 4) {
    fprintf(stderr,"Matrix type descra[0] == %d ",descra[0]);
    fprintf(stderr,"not currently supported in _csrmm.  Supported values: 0,1,3,5\n");
    exit(1);
  }

  if (alpha == 0.0) {
        ScaleArray_TYPE(m, n, c, ldc, beta);
        return;
  }

  if (descra[0] == 1) {
    if (descra[1] == 1) 
    /* Symmetric compressed row matrix stored in lower triangular part */
       CRSymmMatVec_TYPE(m, n, k, alpha, val, indx, pntr, b, ldb, beta, 
                         c, ldc, work, lwork, _SP_base);
    else
    /* Symmetric compressed row matrix stored in lower triangular part */
       CCSymmMatVec_TYPE(m, n, k, alpha, val, indx, pntr, b, ldb, beta, 
                         c, ldc, work, lwork, _SP_base);
  } else {
    if (transa == 1 || transa == 2)
      CompColMatVec_TYPE(m, n, k, alpha, val, indx, pntr, b, ldb, beta, 
                         c, ldc, work, lwork, _SP_base);
    else
      CompRowMatVec_TYPE(m, n, k, alpha, val, indx, pntr, b, ldb, beta, 
                         c, ldc, work, lwork, _SP_base);
  }
}

