/*  -------------------------------------------------------------------  */
/*  NOTE: The _SP_base macro should be defined to indciate the offset    */
/*        of the sparse storage index vectors  (0 or 1)                  */
/*  -------------------------------------------------------------------  */

/*
//                          Sparse BLAS in C
//                             Version 1.1
//       [ Optimization of v. 1.0 for varying vector offsets, plus  ]
//       [   functionality for matvecs with symmetric matrices      ]
//
//    Developed in conjunction with the SparseLib++ Sparse Matrix Library
//
//             National Institute of Standards and Technology, MD.
//                         University of Notre Dame, IN.
//              Authors:  R. Pozo, K. Remington, A. Lumsdaine
//                     (C) 1995 All Rights Reserved
//
//                                NOTICE
//
// Permission to use, copy, modify, and distribute this software and
// its documentation for any purpose and without fee is hereby granted
// provided that the above copyright notice appear in all copies and
// that both the copyright notice and this permission notice appear in
// supporting documentation.
//
// Neither the Authors nor the Institutions (National Institute of Standards
// and Technology, University of Notre Dame) make any representations about
// the suitability of this software for any purpose. This software is
// provided "as is" without expressed or implied warranty.
//
// SparseLib++ was funded in part by the U.S. Department of Energy, the
// National Science Foundation, the University of Notre Dame, and the State
// of Tennessee.
*/

#include <stdlib.h>
#include <stdio.h>
#include "spblas.h"

void CompRow_LU_TYPE(int m, int n, int unitd, const TYPE *dv, 
                 const TYPE alpha, const TYPE *val, 
                 const int *indx, const int *pntr, const TYPE *b, int ldb, 
                 const TYPE beta, TYPE *c, int ldc, TYPE *work, int lwork,
                 const int ind_base)
{
  /*
  // C <-  alpha*D*A^{-1}*B + beta*C
  // C <-  alpha*A^{-1}*D*B + beta*C
  // C is m x n
  // A is m x m  (Sparse; contained in val[] with index arrays indx[] and pntr[])
  // B is m x n
  // D is m x m  (Diagonal; values contained in dv[]0
  // unitd = 1   D = I
  // unitd = 2   left (row scaling)
  // unitd = 3   right (column scaling)
  // alpha is a scalar
  // work is length lwork=m
  */

  int i, j, l;
  TYPE z;

  /*
  // Reference to lwork to avoid compiler warnings
  // (needed in interface, but not currently used)
  */
  if (lwork) { };

  if (alpha == 0.0) {
        ScaleArray_TYPE(m, n, c, ldc, beta);
        return;
  }

  val=val-ind_base;
  indx=indx-ind_base;

  if ( unitd == 1 || unitd == 2) {
    if (n == 1) {
/*    Single triangular system solve, with no right scaling. */
      if ( beta != 0.0 )
         for (i = 0; i < m; i++)
            work[i] = beta*c[i];
      c=c-ind_base;
      for (i = 0; i < m; i++) {
	z = 0;
	for (j = pntr[i]; j < pntr[i+1]; j++)
	  z += c[indx[j]] * val[j];
	c[i+ind_base] = b[i] - z;
      }
      c=c+ind_base;
       if ( unitd == 1)  /* no left scaling */
          if (alpha == 1.0 && beta == 0.0)
             return;
          else if (alpha == 1.0 && beta != 0.0)
             for (j = 0; j < m; j++) c[j] += work[j];
          else if (beta == 0.0)
             for (j = 0; j < m; j++) c[j] *= alpha;
          else
             for (j = 0; j < m; j++) c[j] = alpha*c[j]+work[j];
       else             /* left scaling using elements of dv */
          if (alpha == 1.0 && beta == 0.0)
             for (j = 0; j < m; j++) c[j] *= dv[j];
          else if (alpha == 1.0 && beta != 0.0)
             for (j = 0; j < m; j++) c[j] = dv[j]*c[j]+work[j];
          else if (beta == 0.0)
             for (j = 0; j < m; j++) c[j] *= alpha*dv[j];
          else
             for (j = 0; j < m; j++) c[j] = alpha*dv[j]*c[j]+work[j];
     } else {
/*    Multiple triangular system solves, with no right scaling. */
      for (l = 0; l < n; l++) {
        if ( beta != 0.0 )
           for (i = 0; i < m; i++) 
              work[i] = beta*c[ldc*l+i];
        c=c-ind_base;
	for (i = 0; i < m; i++) {
	  z = 0;
	  for (j = pntr[i]; j < pntr[i+1]; j++)
	    z += c[ldc*l+indx[j]] * val[j];
          c[ldc*l+i+ind_base] = b[ldb*l+i] - z;
        }
       c=c+ind_base;
    
       if ( unitd == 1)  /* no left scaling */
          if (alpha == 1.0 && beta == 0.0)
             return;
          else if (alpha == 1.0 && beta != 0.0)
             for (j = 0; j < m; j++)
               c[ldc*l+j] += work[j];
          else if (beta == 0.0)
             for (j = 0; j < m; j++)
               c[ldc*l+j] *= alpha;
          else
             for (j = 0; j < m; j++)
               c[ldc*l+j] = alpha*c[ldc*l+j]+work[j];
       else             /* left scaling using elements of dv */
          if (alpha == 1.0 && beta == 0.0)
             for (j = 0; j < m; j++)
               c[ldc*l+j] *= dv[j];
          else if (alpha == 1.0 && beta != 0.0)
             for (j = 0; j < m; j++)
               c[ldc*l+j] = dv[j]*c[ldc*l+j]+work[j];
          else if (beta == 0.0)
             for (j = 0; j < m; j++)
               c[ldc*l+j] *= alpha*dv[j];
          else
             for (j = 0; j < m; j++)
               c[ldc*l+j] = alpha*dv[j]*c[ldc*l+j]+work[j];
        
      }
    }
  } else if ( unitd == 3 ) {
    if (n == 1) {
/*    Single triangular system solve, with right scaling. */
      if ( beta != 0.0 )
         for (i = 0; i < m; i++)
            work[i] = beta*c[i];
      c=c-ind_base;
      for (i = 0; i < m; i++) {
        z = 0;
        for (j = pntr[i]; j < pntr[i+1]; j++)
          z += c[indx[j]] * val[j];
        c[i+ind_base] = dv[i]*b[i] - z;
      }
      c=c+ind_base;
      if (alpha == 1.0 && beta == 0.0)
         return;
      else if (alpha == 1.0 && beta != 0.0)
         for (j = 0; j < m; j++) c[j] += work[j];
      else if (beta == 0.0)
         for (j = 0; j < m; j++) c[j] *= alpha;
      else
         for (j = 0; j < m; j++) c[j] = alpha*c[j]+work[j];
    } else {
/*    Multiple triangular system solves, with right scaling. */
      for (l = 0; l < n; l++) {
        if ( beta != 0.0 )
           for (i = 0; i < m; i++)
              work[i] = beta*c[ldc*l+i];
        c=c-ind_base;
        for (i = 0; i < m; i++) {
          z = 0;
          for (j = pntr[i]; j < pntr[i+1]; j++)
            z += c[ldc*l+indx[j]] * val[j];
          c[ldc*l+i+ind_base] = dv[i]*b[ldb*l+i] - z;
        }
        c=c+ind_base;
     
        if (alpha == 1.0 && beta == 0.0)
           return;
        else if (alpha == 1.0 && beta != 0.0)
           for (j = 0; j < m; j++)
             c[ldc*l+j] += work[j];
        else if (beta == 0.0)
           for (j = 0; j < m; j++)
             c[ldc*l+j] *= alpha;
        else
           for (j = 0; j < m; j++)
             c[ldc*l+j] = alpha*c[ldc*l+j]+work[j];
      }
    }
  }
}

/*-----------------------------------------------------------------------------*/
/*-----------------------------------------------------------------------------*/

void CompRow_LD_TYPE(int m, int n, int unitd, const TYPE *dv, 
                 const TYPE alpha, const TYPE *val, 
                 const int *indx, const int *pntr, const TYPE *b, int ldb, 
                 const TYPE beta, TYPE *c, int ldc, TYPE *work, int lwork,
                 const int ind_base)
{
  /*
  // C <-  alpha*D*A^{-1}*B + beta*C
  // C <-  alpha*A^{-1}*D*B + beta*C
  // C is m x n
  // A is m x m  (Sparse; contained in val[] with index arrays indx[] and pntr[])
  // B is m x n
  // D is m x m  (Diagonal; values contained in dv[]0
  // unitd = 1   D = I
  // unitd = 2   left (row scaling)
  // unitd = 3   right (column scaling)
  // alpha is a scalar
  // work is length lwork=m
  */

  int i, j, l;
  TYPE z;

  /*
  // Reference to lwork to avoid compiler warnings
  // (needed in interface, but not currently used)
  */
  if (lwork) { };


  if (alpha == 0.0) {
        ScaleArray_TYPE(m, n, c, ldc, beta);
        return;
  }

  val=val-ind_base;
  indx=indx-ind_base;

  if ( unitd == 1 || unitd == 2) {
    if (n == 1) {
/*    Single triangular system solve, with no right scaling. */
      if ( beta != 0.0 )
         for (i = 0; i < m; i++)
            work[i] = beta*c[i];
      c=c-ind_base;
      for (i = 0; i < m; i++) {
	z = 0;
	for (j = pntr[i]; j < pntr[i+1] - 1; j++)
	  z += c[indx[j]] * val[j];
	c[i+ind_base] = (b[i] - z) / val[pntr[i+1]-1];
      }
      c=c+ind_base;
       if ( unitd == 1)  /* no left scaling */
          if (alpha == 1.0 && beta == 0.0)
             return;
          else if (alpha == 1.0 && beta != 0.0)
             for (j = 0; j < m; j++) c[j] += work[j];
          else if (beta == 0.0)
             for (j = 0; j < m; j++) c[j] *= alpha;
          else
             for (j = 0; j < m; j++) c[j] = alpha*c[j]+work[j];
       else             /* left scaling using elements of dv */
          if (alpha == 1.0 && beta == 0.0)
             for (j = 0; j < m; j++) c[j] *= dv[j];
          else if (alpha == 1.0 && beta != 0.0)
             for (j = 0; j < m; j++) c[j] = dv[j]*c[j]+work[j];
          else if (beta == 0.0)
             for (j = 0; j < m; j++) c[j] *= alpha*dv[j];
          else
             for (j = 0; j < m; j++) c[j] = alpha*dv[j]*c[j]+work[j];
     } else {
/*    Multiple triangular system solves, with no right scaling. */
      for (l = 0; l < n; l++) {
        if ( beta != 0.0 )
           for (i = 0; i < m; i++) 
              work[i] = beta*c[ldc*l+i];
        c=c-ind_base;
	for (i = 0; i < m; i++) {
	  z = 0;
	  for (j = pntr[i]; j < pntr[i+1] - 1; j++)
	    z += c[ldc*l+indx[j]] * val[j];
	     c[ldc*l+i+ind_base] = 
	              (b[ldb*l+i] - z) / val[pntr[i+1]-1];
       }
       c=c+ind_base;
       if ( unitd == 1)  /* no left scaling */
          if (alpha == 1.0 && beta == 0.0)
             return;
          else if (alpha == 1.0 && beta != 0.0)
             for (j = 0; j < m; j++)
               c[ldc*l+j] += work[j];
          else if (beta == 0.0)
             for (j = 0; j < m; j++)
               c[ldc*l+j] *= alpha;
          else
             for (j = 0; j < m; j++)
               c[ldc*l+j] = alpha*c[ldc*l+j]+work[j];
       else             /* left scaling using elements of dv */
          if (alpha == 1.0 && beta == 0.0)
             for (j = 0; j < m; j++)
               c[ldc*l+j] *= dv[j];
          else if (alpha == 1.0 && beta != 0.0)
             for (j = 0; j < m; j++)
               c[ldc*l+j] = dv[j]*c[ldc*l+j]+work[j];
          else if (beta == 0.0)
             for (j = 0; j < m; j++)
               c[ldc*l+j] *= alpha*dv[j];
          else
             for (j = 0; j < m; j++)
               c[ldc*l+j] = alpha*dv[j]*c[ldc*l+j]+work[j];
        
      }
    }
  } else if ( unitd == 3 ) {
    if (n == 1) {
/*    Single triangular system solve, with right scaling. */
      if ( beta != 0.0 )
         for (i = 0; i < m; i++)
            work[i] = beta*c[i];
      c=c-ind_base;
      for (i = 0; i < m; i++) {
        z = 0;
        for (j = pntr[i]; j < pntr[i+1] - 1; j++)
          z += c[indx[j]] * val[j];
        c[i+ind_base] = (dv[i]*b[i] - z) / val[pntr[i+1]-1];
      }
      c=c+ind_base;
      if (alpha == 1.0 && beta == 0.0)
         return;
      else if (alpha == 1.0 && beta != 0.0)
         for (j = 0; j < m; j++) c[j] += work[j];
      else if (beta == 0.0)
         for (j = 0; j < m; j++) c[j] *= alpha;
      else
         for (j = 0; j < m; j++) c[j] = alpha*c[j]+work[j];
    } else {
/*    Multiple triangular system solves, with right scaling. */
      for (l = 0; l < n; l++) {
        if ( beta != 0.0 )
           for (i = 0; i < m; i++)
              work[i] = beta*c[ldc*l+i];
        c=c-ind_base;
        for (i = 0; i < m; i++) {
          z = 0;
          for (j = pntr[i]; j < pntr[i+1] - 1; j++)
            z += c[ldc*l+indx[j]] * val[j];
             c[ldc*l+i+ind_base] =
                      (dv[i]*b[ldb*l+i] - z) / val[pntr[i+1]-1];
        }
        c=c+ind_base;
     
        if (alpha == 1.0 && beta == 0.0)
           return;
        else if (alpha == 1.0 && beta != 0.0)
           for (j = 0; j < m; j++)
             c[ldc*l+j] += work[j];
        else if (beta == 0.0)
           for (j = 0; j < m; j++)
             c[ldc*l+j] *= alpha;
        else
           for (j = 0; j < m; j++)
             c[ldc*l+j] = alpha*c[ldc*l+j]+work[j];
      }
    }
  }
}

/*-----------------------------------------------------------------------------*/
/*-----------------------------------------------------------------------------*/

void CompRow_UU_TYPE(int m, int n, int unitd, const TYPE *dv, 
                 const TYPE alpha, const TYPE *val, 
                 const int *indx, const int *pntr, const TYPE *b, int ldb, 
                 const TYPE beta, TYPE *c, int ldc, TYPE *work, int lwork,
                 const int ind_base)
{
  /*
  // C <-  alpha*D*A^{-1}*B + beta*C
  // C <-  alpha*A^{-1}*D*B + beta*C
  // C is m x n
  // A is m x m  (Sparse; contained in val[] with index arrays indx[] and pntr[])
  // B is m x n
  // D is m x m  (Diagonal; values contained in dv[]0
  // unitd = 1   D = I
  // unitd = 2   left (row scaling)
  // unitd = 3   right (column scaling)
  // alpha is a scalar
  // work is length lwork=m
  */

  int i, j, l;
  TYPE z;

  /*
  // Reference to lwork to avoid compiler warnings
  // (needed in interface, but not currently used)
  */
  if (lwork) { };

  if (alpha == 0.0) {
        ScaleArray_TYPE(m, n, c, ldc, beta);
        return;
  }

  val=val-ind_base;
  indx=indx-ind_base;

  if ( unitd == 1 || unitd == 2) {
    if (n == 1) {
/*    Single triangular system solve, with no right scaling. */
      if ( beta != 0.0 )
         for (i = 0; i < m; i++)
            work[i] = beta*c[i];
      c=c-ind_base;
      for (i = m - 1; i >= 0; i--) {
        z = 0;
        for (j = pntr[i]; j < pntr[i+1]; j++)
          z += c[indx[j]] * val[j];
        c[i+ind_base] = b[i] - z;
      }
      c=c+ind_base;
       if ( unitd == 1)  /* no left scaling */
          if (alpha == 1.0 && beta == 0.0)
             return;
          else if (alpha == 1.0 && beta != 0.0)
             for (j = 0; j < m; j++) c[j] += work[j];
          else if (beta == 0.0)
             for (j = 0; j < m; j++) c[j] *= alpha;
          else
             for (j = 0; j < m; j++) c[j] = alpha*c[j]+work[j];
       else             /* left scaling using elements of dv */
          if (alpha == 1.0 && beta == 0.0)
             for (j = 0; j < m; j++) c[j] *= dv[j];
          else if (alpha == 1.0 && beta != 0.0)
             for (j = 0; j < m; j++) c[j] = dv[j]*c[j]+work[j];
          else if (beta == 0.0)
             for (j = 0; j < m; j++) c[j] *= alpha*dv[j];
          else
             for (j = 0; j < m; j++) c[j] = alpha*dv[j]*c[j]+work[j];
     } else {
/*    Multiple triangular system solves, with no right scaling. */
      for (l = 0; l < n; l++) {
        if ( beta != 0.0 )
           for (i = 0; i < m; i++) 
              work[i] = beta*c[ldc*l+i];
        c=c-ind_base;
        for (i = m - 1; i >= 0; i--) {
          z = 0;
          for (j = pntr[i]; j < pntr[i+1]; j++)
            z += c[ldc*l+indx[j]] * val[j];
          c[ldc*l+i+ind_base] = (b[ldb*l+i] - z);
        }
       c=c+ind_base;
       if ( unitd == 1)  /* no left scaling */
          if (alpha == 1.0 && beta == 0.0)
             return;
          else if (alpha == 1.0 && beta != 0.0)
             for (j = 0; j < m; j++)
               c[ldc*l+j] += work[j];
          else if (beta == 0.0)
             for (j = 0; j < m; j++)
               c[ldc*l+j] *= alpha;
          else
             for (j = 0; j < m; j++)
               c[ldc*l+j] = alpha*c[ldc*l+j]+work[j];
       else             /* left scaling using elements of dv */
          if (alpha == 1.0 && beta == 0.0)
             for (j = 0; j < m; j++)
               c[ldc*l+j] *= dv[j];
          else if (alpha == 1.0 && beta != 0.0)
             for (j = 0; j < m; j++)
               c[ldc*l+j] = dv[j]*c[ldc*l+j]+work[j];
          else if (beta == 0.0)
             for (j = 0; j < m; j++)
               c[ldc*l+j] *= alpha*dv[j];
          else
             for (j = 0; j < m; j++)
               c[ldc*l+j] = alpha*dv[j]*c[ldc*l+j]+work[j];
        
      }
    }
  } else if ( unitd == 3 ) {
    if (n == 1) {
/*    Single triangular system solve, with right scaling. */
      if ( beta != 0.0 )
         for (i = 0; i < m; i++)
            work[i] = beta*c[i];
      c=c-ind_base;
      for (i = m - 1; i >= 0; i--) {
        z = 0;
        for (j = pntr[i]; j < pntr[i+1]; j++)
          z += c[indx[j]] * val[j];
        c[i+ind_base] = dv[i]*b[i] - z;
      }
      c=c+ind_base;
      if (alpha == 1.0 && beta == 0.0)
         return;
      else if (alpha == 1.0 && beta != 0.0)
         for (j = 0; j < m; j++) c[j] += work[j];
      else if (beta == 0.0)
         for (j = 0; j < m; j++) c[j] *= alpha;
      else
         for (j = 0; j < m; j++) c[j] = alpha*c[j]+work[j];
    } else {
/*    Multiple triangular system solves, with right scaling. */
      for (l = 0; l < n; l++) {
        if ( beta != 0.0 )
           for (i = 0; i < m; i++)
              work[i] = beta*c[ldc*l+i];
        c=c-ind_base;
        for (i = m - 1; i >= 0; i--) {
          z = 0;
          for (j = pntr[i]; j < pntr[i+1]; j++)
            z += c[ldc*l+indx[j]] * val[j];
             c[ldc*l+i+ind_base] =
                      dv[i]*b[ldb*l+i] - z;
        }
        c=c+ind_base;
        if (alpha == 1.0 && beta == 0.0)
           return;
        else if (alpha == 1.0 && beta != 0.0)
           for (j = 0; j < m; j++)
             c[ldc*l+j] += work[j];
        else if (beta == 0.0)
           for (j = 0; j < m; j++)
             c[ldc*l+j] *= alpha;
        else
           for (j = 0; j < m; j++)
             c[ldc*l+j] = alpha*c[ldc*l+j]+work[j];
      }
    }
  }
}

/*-----------------------------------------------------------------------------*/
/*-----------------------------------------------------------------------------*/

void CompRow_UD_TYPE(int m, int n, int unitd, const TYPE *dv, 
                 const TYPE alpha, const TYPE *val, 
                 const int *indx, const int *pntr, const TYPE *b, int ldb, 
                 const TYPE beta, TYPE *c, int ldc, TYPE *work, int lwork,
                 const int ind_base)
{
  /*
  // C <-  alpha*D*A^{-1}*B + beta*C
  // C <-  alpha*A^{-1}*D*B + beta*C
  // C is m x n
  // A is m x m  (Sparse; contained in val[] with index arrays indx[] and pntr[])
  // B is m x n
  // D is m x m  (Diagonal; values contained in dv[]0
  // unitd = 1   D = I
  // unitd = 2   left (row scaling)
  // unitd = 3   right (column scaling)
  // alpha is a scalar
  // work is length lwork=m
  */

  int i, j, l;
  TYPE z;

  /*
  // Reference to lwork to avoid compiler warnings
  // (needed in interface, but not currently used)
  */
  if (lwork) { };

  if (alpha == 0.0) {
        ScaleArray_TYPE(m, n, c, ldc, beta);
        return;
  }

  val=val-ind_base;
  indx=indx-ind_base;

  if ( unitd == 1 || unitd == 2) {
    if (n == 1) {
/*    Single triangular system solve, with no right scaling. */
      if ( beta != 0.0 )
         for (i = 0; i < m; i++)
            work[i] = beta*c[i];
      c=c-ind_base;
      for (i = m - 1; i >= 0; i--) {
        z = 0;
        for (j = pntr[i]+1; j < pntr[i+1]; j++)
          z += c[indx[j]] * val[j];
	c[i+ind_base] = (b[i] - z) / val[pntr[i]];
      }
      c=c+ind_base;
       if ( unitd == 1)  /* no left scaling */
          if (alpha == 1.0 && beta == 0.0)
             return;
          else if (alpha == 1.0 && beta != 0.0)
             for (j = 0; j < m; j++) c[j] += work[j];
          else if (beta == 0.0)
             for (j = 0; j < m; j++) c[j] *= alpha;
          else
             for (j = 0; j < m; j++) c[j] = alpha*c[j]+work[j];
       else             /* left scaling using elements of dv */
          if (alpha == 1.0 && beta == 0.0)
             for (j = 0; j < m; j++) c[j] *= dv[j];
          else if (alpha == 1.0 && beta != 0.0)
             for (j = 0; j < m; j++) c[j] = dv[j]*c[j]+work[j];
          else if (beta == 0.0)
             for (j = 0; j < m; j++) c[j] *= alpha*dv[j];
          else
             for (j = 0; j < m; j++) c[j] = alpha*dv[j]*c[j]+work[j];
     } else {
/*    Multiple triangular system solves, with no right scaling. */
      for (l = 0; l < n; l++) {
        if ( beta != 0.0 )
           for (i = 0; i < m; i++) 
              work[i] = beta*c[ldc*l+i];
        c=c-ind_base;
        for (i = m - 1; i >= 0; i--) {
          z = 0;
          for (j = pntr[i]+1; j < pntr[i+1]; j++)
            z += c[ldc*l+indx[j]] * val[j];
	    c[ldc*l+i+ind_base] = (b[ldb*l+i] - z) / val[pntr[i]];
        }
       c=c+ind_base;
       if ( unitd == 1)  /* no left scaling */
          if (alpha == 1.0 && beta == 0.0)
             return;
          else if (alpha == 1.0 && beta != 0.0)
             for (j = 0; j < m; j++)
               c[ldc*l+j] += work[j];
          else if (beta == 0.0)
             for (j = 0; j < m; j++)
               c[ldc*l+j] *= alpha;
          else
             for (j = 0; j < m; j++)
               c[ldc*l+j] = alpha*c[ldc*l+j]+work[j];
       else             /* left scaling using elements of dv */
          if (alpha == 1.0 && beta == 0.0)
             for (j = 0; j < m; j++)
               c[ldc*l+j] *= dv[j];
          else if (alpha == 1.0 && beta != 0.0)
             for (j = 0; j < m; j++)
               c[ldc*l+j] = dv[j]*c[ldc*l+j]+work[j];
          else if (beta == 0.0)
             for (j = 0; j < m; j++)
               c[ldc*l+j] *= alpha*dv[j];
          else
             for (j = 0; j < m; j++)
               c[ldc*l+j] = alpha*dv[j]*c[ldc*l+j]+work[j];
        
      }
    }
  } else if ( unitd == 3 ) {
    if (n == 1) {
/*    Single triangular system solve, with right scaling. */
      if ( beta != 0.0 )
         for (i = 0; i < m; i++)
            work[i] = beta*c[i];
      c=c-ind_base;
      for (i = m - 1; i >= 0; i--) {
        z = 0;
        for (j = pntr[i]+1; j < pntr[i+1]; j++)
          z += c[indx[j]] * val[j];
        c[i+ind_base] = (dv[i]*b[i] - z)/ val[pntr[i]];
      }
      c=c+ind_base;
      if (alpha == 1.0 && beta == 0.0)
         return;
      else if (alpha == 1.0 && beta != 0.0)
         for (j = 0; j < m; j++) c[j] += work[j];
      else if (beta == 0.0)
         for (j = 0; j < m; j++) c[j] *= alpha;
      else
         for (j = 0; j < m; j++) c[j] = alpha*c[j]+work[j];
    } else {
/*    Multiple triangular system solves, with right scaling. */
      for (l = 0; l < n; l++) {
        if ( beta != 0.0 )
           for (i = 0; i < m; i++)
              work[i] = beta*c[ldc*l+i];
        c=c-ind_base;
        for (i = m - 1; i >= 0; i--) {
          z = 0;
          for (j = pntr[i]+1; j < pntr[i+1]; j++)
            z += c[ldc*l+indx[j]] * val[j];
             c[ldc*l+i+ind_base] = (dv[i]*b[ldb*l+i] - z)/ val[pntr[i]];
        }
        c=c+ind_base;
        if (alpha == 1.0 && beta == 0.0)
           return;
        else if (alpha == 1.0 && beta != 0.0)
           for (j = 0; j < m; j++)
             c[ldc*l+j] += work[j];
        else if (beta == 0.0)
           for (j = 0; j < m; j++)
             c[ldc*l+j] *= alpha;
        else
           for (j = 0; j < m; j++)
             c[ldc*l+j] = alpha*c[ldc*l+j]+work[j];
      }
    }
  }
}

/*-----------------------------------------------------------------------------*/
/*-----------------------------------------------------------------------------*/

/* F77 interface routine: */
void _csrsm(const int transa, const int m, const int n,
            const int unitd, const TYPE *dv, const TYPE alpha,
            const int descra[], const TYPE *val,
            const int *indx, const int *pntr, const TYPE *b, int ldb,
            const TYPE beta, TYPE *c, const int ldc,
            TYPE *work, const int lwork)
{
  /*
  // C <-  alpha*D*A^{-1}*B + beta*C
  // C <-  alpha*A^{-1}*D*B + beta*C
  // C is m x n
  // A is m x m  (Sparse; contained in val[] with index arrays indx[] and pntr[])
  // B is m x n
  // D is m x m  (Diagonal; values contained in dv[]0
  // unitd = 1   D = I
  // unitd = 2   left (row scaling)
  // unitd = 3   right (column scaling)
  // alpha is a scalar
  */

  if (descra[0] != 3) {
    fprintf(stderr,"Must have triangular matrix for _csrsm. (descra[0] == 3)\n");
    exit(1);
  }

  if (transa == 0) {
    if (descra[1] == 1) {
      if (descra[2] == 0)
	CompRow_LD_TYPE(m, n, unitd, dv, alpha, val, 
                         indx, pntr, b, ldb, beta, c, ldc, 
                         work, lwork, _SP_base);
      else if (descra[2] == 1)
	CompRow_LU_TYPE(m, n, unitd, dv, alpha, val, 
                         indx, pntr, b, ldb, beta, c, ldc,
                         work, lwork, _SP_base);
      else {
	fprintf(stderr,"Bad descra[2] in _csrsm.\n");
	exit(1);
      }
    } else if (descra[1] == 2) {
      if (descra[2] == 0)
	CompRow_UD_TYPE(m, n, unitd, dv, alpha, val, 
                         indx, pntr, b, ldb, beta, c, ldc,
                         work, lwork, _SP_base);
      else if (descra[2] == 1)
	CompRow_UU_TYPE(m, n, unitd, dv, alpha, val, 
                         indx, pntr, b, ldb, beta, c, ldc,
                         work, lwork, _SP_base);
      else {
	fprintf(stderr,"Bad descra[2] in _csrsm.\n");
	exit(1);
      }
    }
  } else if (transa == 1 || transa == 2) {
    if (descra[1] == 1) {
      if (descra[2] == 0)
	CompCol_UD_TYPE(m, n, unitd, dv, alpha, val, 
                         indx, pntr, b, ldb, beta, c, ldc,
                         work, lwork, _SP_base);
      else if (descra[2] == 1)
	CompCol_UU_TYPE(m, n, unitd, dv, alpha, val, 
                         indx, pntr, b, ldb, beta, c, ldc,
                         work, lwork, _SP_base);
      else {
	fprintf(stderr,"Bad descra[2] in _csrsm.\n");
	exit(1);
      }
    } else if (descra[1] == 2) {
      if (descra[2] == 0)
	CompCol_LD_TYPE(m, n, unitd, dv, alpha, val, 
                         indx, pntr, b, ldb, beta, c, ldc,
                         work, lwork, _SP_base);
      else if (descra[2] == 1)
	CompCol_LU_TYPE(m, n, unitd, dv, alpha, val, 
                         indx, pntr, b, ldb, beta, c, ldc,
                         work, lwork, _SP_base);
      else {
	fprintf(stderr,"Bad descra[2] in _csrsm.\n");
	exit(1);
      }
    } else {
      fprintf(stderr,"Bad descra[1] in _csrsm.\n");
      exit(1);
    }
  } else {
    fprintf(stderr,"Bad transa in _csrsm.\n");
    exit(1);
  }
}

