/*  -------------------------------------------------------------------  */
/*  NOTE: The _SP_base macro should be defined to indicate the offset    */
/*        of the sparse storage index vectors  (0 or 1)                  */
/*  -------------------------------------------------------------------  */

/*
//                          Sparse BLAS in C
//                             Version 1.1
//       [ Optimization of v. 1.0 for varying vector offsets, plus  ]
//       [   functionality for matvecs with symmetric matrices      ]
//
//    Developed in conjunction with the SparseLib++ Sparse Matrix Library
//
//             National Institute of Standards and Technology, MD.
//                         University of Notre Dame, IN.
//              Authors:  R. Pozo, K. Remington, A. Lumsdaine
//                     (C) 1995 All Rights Reserved
//
//                                NOTICE
//
// Permission to use, copy, modify, and distribute this software and
// its documentation for any purpose and without fee is hereby granted
// provided that the above copyright notice appear in all copies and
// that both the copyright notice and this permission notice appear in
// supporting documentation.
//
// Neither the Authors nor the Institutions (National Institute of Standards
// and Technology, University of Notre Dame) make any representations about
// the suitability of this software for any purpose. This software is
// provided "as is" without expressed or implied warranty.
//
// SparseLib++ was funded in part by the U.S. Department of Energy, the
// National Science Foundation, the University of Notre Dame, and the State
// of Tennessee.
*/

#include <stdlib.h>
#include <stdio.h>
#include "spblas.h"

void CompColMatVec_float(int m, int n, int k, const float alpha,
		   const float *val, const int *indx, const int *pntr,
		   const float *b, int ldb, const float beta, float *c, int ldc,
                   float *work, const int lwork, const int ind_base)
{
  /*
  // C <- alpha*A*B + beta*C
  // C is m x n (Dense)
  // A is m x k (Sparse; contained in val[] with index arrays indx[] and jndx[])
  // B is k x n (Dense)
  // alpha and beta are scalars
  */

  int i, j, l;

  /* Reference to m and lwork to avoid compiler warnings */
  /* (needed in interface, but not currently used)       */
  if (m&&lwork) { };

  if (alpha == 0.0) {
     ScaleArray_float(m, n, c, ldc, beta);
     return;
  }

  val=val-ind_base;
  indx=indx-ind_base;

  if (alpha == 1.0) {
    if (n == 1) {
/*    Just one matrix * vector, with no scaling:   */
      if (beta != 0 ) for (i=0;i<m;i++) work[i] = beta*c[i];
      for (i=0;i<m;i++) c[i] = 0;
      c=c-ind_base;
      for (i = 0; i < k; i++) 
	for (j = pntr[i]; j < pntr[i+1]; j++)
	  c[indx[j]] += b[i] * val[j];
      c=c+ind_base;
      if (beta != 0 ) for (i=0;i<m;i++) c[i] += work[i];
    } else {
/*    Multiple matrix * vector, with no scaling:   */
      for (l = 0; l < n; l++){
        if (beta != 0 ) 
            for (i=0;i<m;i++) 
               work[i] = beta*c[ldc*l+i];
        for (i=0;i<m;i++) 
            c[ldc*l+i] = 0;
        c=c-ind_base;
	for (i = 0; i < k; i++)
	  for (j = pntr[i]; j < pntr[i+1]; j++)
	    c[ldc*l+indx[j]] += b[ldb*l+i] * val[j];
        c=c+ind_base;
        if (beta != 0 ) 
            for (i=0;i<m;i++) 
               c[ldc*l+i] += work[i];
      }
    }
  } else {
    if (n == 1) {
/*    Just one matrix * vector, with scaling:    */
      if (beta != 0 ) for (i=0;i<m;i++) work[i] = beta*c[i];
      for (i=0;i<m;i++) c[i] = 0;
      c=c-ind_base;
      for (i = 0; i < k; i++)
	for (j = pntr[i]; j < pntr[i+1]; j++)
	  c[indx[j]] += alpha * b[i] * val[j];
      c=c+ind_base;
      if (beta != 0 ) for (i=0;i<m;i++) c[i] += work[i];
    } else {
/*    Multiple matrix * vector, with scaling:    */
      for (l = 0; l < n; l++) {
        if (beta != 0 ) 
            for (i=0;i<m;k++) 
               work[i] = beta*c[ldc*l+i];
        for (i=0;i<m;i++) 
            c[ldc*l+i] = 0;
        c=c-ind_base;
	for (i = 0; i < k; i++)
	  for (j = pntr[i]; j < pntr[i+1]; j++)
	    c[ldc*l+indx[j]] +=
	      alpha * b[ldb*l+i] * val[j];
        c=c+ind_base;
        if (beta != 0 ) 
            for (i=0;i<m;i++) 
               c[ldc*l+i] += work[i];
      }
    }
  }
}

void CCSymmMatVec_float(int m, int n, int k, const float alpha,
		   const float *val, const int *indx, const int *pntr,
		   const float *b, int ldb, const float beta, float *c, int ldc,
                   float *work, const int lwork, const int ind_base)
{
  /*
  // C <- alpha*A*B + beta*C
  // C is m x n (Dense)
  // A is m x k (Symmetric Sparse; contained in val[] with index arrays indx[] and jndx[])
  // B is k x n (Dense)
  // alpha and beta are scalars
  */

  int i, j, l;
  int col_ptrlow, col_ptrhi;

  /* Reference to m and lwork to avoid compiler warnings */
  /* (needed in interface, but not currently used)       */
  if (m&&lwork) { };

  if (alpha == 0.0) {
     ScaleArray_float(m, n, c, ldc, beta);
     return;
  }

  val=val-ind_base;
  indx=indx-ind_base;

  if (alpha == 1.0) {
    if (n == 1) {
/*    Just one matrix * vector, with no scaling:   */
      if (beta != 0 ) for (i=0;i<m;i++) work[i] = beta*c[i];
      for (i=0;i<m;i++) c[i] = 0;
      for(j=0;j<k;j++){
          col_ptrlow = pntr[j];
          col_ptrhi = pntr[j+1];
          if ( indx[col_ptrlow] == j+ind_base) {
             c[j] += b[j]*val[col_ptrlow];
             col_ptrlow++;
          }
          b=b-ind_base;
          for(i=col_ptrlow;i<col_ptrhi;i++){
             c[j] += b[indx[i]]*val[i];
          }
          c=c-ind_base;
          b=b+ind_base;
          for(i=col_ptrlow;i<col_ptrhi;i++){
             c[indx[i]]+= b[j]*val[i];
          }
          if (_SP_base == 1) c++;
       }
      if (beta != 0 ) for (i=0;i<m;i++) c[i] += work[i];
    } else {
/*    Multiple matrix * vector, with no scaling:   */
      for (l = 0; l < n; l++){
        if (beta != 0 ) 
            for (i=0;i<m;i++) work[i] = beta*c[ldc*l+i];
        for (i=0;i<m;i++) c[ldc*l+i] = 0;
        for(j=0;j<k;j++){
          col_ptrlow = pntr[j];
          col_ptrhi = pntr[j+1];
          if ( indx[col_ptrlow] == j+ind_base ) {
	     c[ldc*l+j] += b[ldb*l+j] * val[col_ptrlow];
             col_ptrlow++;
          }
          b=b-ind_base;
          for(i=col_ptrlow;i<col_ptrhi;i++){
	     c[ldc*l+j] += b[ldb*l+indx[i]] * val[i];
          }
          c=c-ind_base;
          b=b+ind_base;
          for(i=col_ptrlow;i<col_ptrhi;i++){
	     c[ldc*l+indx[i]] += b[ldb*l+j] * val[i];
          }
          c=c+ind_base;
        }
        if (beta != 0 ) 
            for (i=0;i<m;i++) c[ldc*l+i] += work[i];
      }
    }
  } else {
    if (n == 1) {
/*    Just one matrix * vector, with scaling:    */
      if (beta != 0 ) for (i=0;i<m;i++) work[i] = beta*c[i];
      for (i=0;i<m;i++) c[i] = 0;
      for(j=0;j<k;j++){
          col_ptrlow = pntr[j];
          col_ptrhi = pntr[j+1];
          if ( indx[col_ptrlow] == j+ind_base ) {
             c[j] += alpha*b[j]*val[col_ptrlow];
             col_ptrlow++;
          }
          b=b-ind_base;
          for(i=col_ptrlow;i<col_ptrhi;i++){
             c[j] += alpha*b[indx[i]]*val[i];
          }
          c=c-ind_base;
          b=b+ind_base;
          for(i=col_ptrlow;i<col_ptrhi;i++){
             c[indx[i]]+= alpha*b[j]*val[i];
          }
          c=c+ind_base;
       }
      if (beta != 0 ) for (i=0;i<m;i++) c[i] += work[i];
    } else {
/*    Multiple matrix * vector, with scaling:    */
      for (l = 0; l < n; l++) {
        if (beta != 0 ) for (i=0;i<m;k++) work[i] = beta*c[ldc*l+i];
        for (i=0;i<m;i++) c[ldc*l+i] = 0;
        for(j=0;j<k;j++){
          col_ptrlow = pntr[j];
          col_ptrhi = pntr[j+1];
          if ( indx[col_ptrlow] == j+ind_base ) {
	     c[ldc*l+j] += alpha*b[ldb*l+j] * val[col_ptrlow];
             col_ptrlow++;
          }
          b=b-ind_base;
          for(i=col_ptrlow;i<col_ptrhi;i++){
	     c[ldc*l+j] += alpha*b[ldb*l+indx[i]] * val[i];
          }
          c=c-ind_base;
          b=b+ind_base;
          for(i=col_ptrlow;i<col_ptrhi;i++){
	     c[ldc*l+indx[i]] += alpha*b[ldb*l+j] * val[i];
          }
          c=c+ind_base;
        }
        if (beta != 0 ) for (i=0;i<m;i++) c[ldc*l+i] += work[i];
      }
    }
  }
}
/*-----------------------------------------------------------------------------*/
/*-----------------------------------------------------------------------------*/

/* F77 interface routine: */
void scscmm(const int transa, const int m, const int n, const int k,
            const float alpha, const int descra[], const float *val,
            const int *indx, const int *pntr, const float *b, int ldb,
            const float beta, float *c, const int ldc,
            float *work, const int lwork)
{
/*-----------------------------------------------------------------------------*/
/*-----------------------------------------------------------------------------*/
/* F77 interface:
 * scscmm -- comp sparse column matrix-matrix multiply
 *
 * Arguments:
 *
 * int &transa	Indicates how to operate with the sparse matrix
 *		0 : operate with matrix
 *		1 : operate with transpose matrix
 *		2 : operate with conjugate transpose matrix
 *
 * int &m	Number of rows in matrix c
 *
 * int &n	Number of columns in matrix c
 *
 * int &k	Number of rows in matrix b
 *
 * float		&alpha Scalar parameter
 *
 * float		&beta Scalar parameter
 *
 * int descra[]	Descriptor argument.  Nine element integer array
 *		descra[0] matrix structure
 *			0 : general
 *			1 : symmetric
 *			2 : Hermition
 *			3 : Triangular
 *			4 : Anti-Symmetric
 *			5 : Diagonal
 *		descra[1] upper/lower triangular indicator
 *			1 : lower
 *			2 : upper
 *		descra[2] main diagonal type
 *			0 : non-unit
 *			1 : unit
 *
 * float		*val	scalar array of length nnz containing matrix entries
 *
 * int *indx	integer array of length nnz containing row indices
 *
 * int *pntr	integer array of length k+1 such that pntr(j)-pntr(1)
 *		points to location in val of the first element in column j
 *
 * float		*b	rectangular array with first dimension ldb
 *
 * float		*c	rectangular array with first dimension ldc
 *
 * float		*work	scratch array of length lwork.  lwork should be at least
 *		max(m,n)
 *
 */
/*-----------------------------------------------------------------------------*/
/*-----------------------------------------------------------------------------*/
  if (descra[0] == 2 || descra[0] == 4) {
    fprintf(stderr,"Matrix type descra[0] == %d ",descra[0]);
    fprintf(stderr,"not currently supported in scscmm.  Supported values: 0,1,3,5\n");
    exit(1);
  }

  if (alpha == 0.0) {
        ScaleArray_float(m, n, c, ldc, beta);
        return;
  }
 
  if (descra[0] == 1 ) {
    if (descra[1] == 1 ) 
    /* Symmetric compressed column matrix stored in lower triangular part */
       CCSymmMatVec_float(m, n, k, alpha, val, indx, pntr, b, ldb, beta, 
                         c, ldc, work, lwork, _SP_base);
    else
    /* Symmetric compressed column matrix stored in upper triangular part */
       CRSymmMatVec_float(m, n, k, alpha, val, indx, pntr, b, ldb, beta, 
                         c, ldc, work, lwork, _SP_base);
  } else {
    if (transa == 1 || transa == 2)
      CompRowMatVec_float(m, n, k, alpha, val, indx, pntr, b, ldb, beta, 
                         c, ldc, work, lwork, _SP_base);
    else
      CompColMatVec_float(m, n, k, alpha, val, indx, pntr, b, ldb, beta,
                         c, ldc, work, lwork, _SP_base);
  }
}

