/*  -------------------------------------------------------------------  */
/*  NOTE: The _SP_base macro should be defined to indciate the offset    */
/*        of the sparse storage index vectors  (0 or 1)                  */
/*  -------------------------------------------------------------------  */

#include <stdlib.h>
#include "compcol_double.h"
#include "comprow_double.h"
#include "ilupre_double.h"
#include "icpre_double.h"
#include "iohb_double.h"
#include "spblas.h"

int main(int argc, char * argv[])
{
/***************************************************************/
//  Test the following functions from the Sparse BLAS:         //
//       CompRow_LowerDiagSolve()                              //
//       CompRow_UpperDiagSolve()                              //
//       CompCol_LowerDiagSolve()                              //
//       CompCol_UpperDiagSolve()                              //
//                                                             //
//  C <-- alpha*D*inv(A)*B + beta*C  (row scaling)             //
//  C <-- alpha*inv(A)*D*B + beta*C  (column scaling)          //
//                                                             //
//  where C and B have more than one column.                   //
//                                                             //
//  This program takes two arguments, specifying the scalar    //
//  constants alpha and beta.                                  //
//  The underlying code takes different paths if:              //
//          alpha == 0.0 or 1.0                                //
//      or  beta  == 0.0 or 1.0                                //
//  Therefore, the testing program should be run with various  //
//  combinations of these parameters.                          //
//                                                             //
/***************************************************************/
//                                                             //
//                   [    1  0  0  0  0   ]                    //
//                                                             //
//                   |    2  3  0  0  0   |                    //
//                                                             //
//                   |    0  4  5  0  0   |                    //
//                                                             //
//                   |    0  0  6  7  0   |                    //
//                                                             //
//                   [    8  0  0  0  9   ]                    //
//                                                             //
/***************************************************************/

  int verbose = 0;

  if (argc > 4 || argc < 3) {
     cerr << "Usage: " << argv[0] << " alpha beta [-v]" << endl;
     exit(-1);
  }
  if (argc > 3) {
     if (argv[3][1] == 'v') verbose = 1;
  } 

  if (verbose) cout << "Testing Sparse BLAS" << endl;

  double tolerance = 1.e-12;
  int errcount = 0;
  int i;

//Column storage:
  double cval[]= {1.0, 2.0, 8.0, 3.0, 4.0, 5.0, 6.0, 7.0, 9.0};
  int rowind[] = {1,   2,   5,   2,   3,   3,   4,   4,   5   };
  int colptr[] = {1,             4,        6,        8,   9,      10};
//Row storage:
  double rval[]= {1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0};
  int colind[] = {1,   1,   2,   2,   3,   3,   4,   1,   5   };
  int rowptr[] = {1,   2,        4,        6,   8,                10};
//Coordinate storage:
  int indx[] = {1,   2,   5,   2,   3,   3,   4,   4,   5   };
  int jndx[] = {1,   1,   1,   2,   2,   3,   3,   4,   5   };
//Decrement indices and pointers if internal storage scheme is 0-based:
  if ( _SP_base != 1) {
    for (i=0;i<9;i++) {
       rowind[i]--;
       colind[i]--;
       indx[i]--;
       jndx[i]--;
    }
    for (i=0;i<6;i++) {
       colptr[i]--;
       rowptr[i]--;
    }
  }
  CompRow_Mat_double Ar(5,5,9,rval,rowptr,colind);
  CompCol_Mat_double Ac(Ar);

  int m = Ar.dim(0);
  
  VECTOR_double b(m,0.0), x(m), dv(m), ytrue(m), work(m);
  VECTOR_double b2(2*m,0.0), x2(2*m), y2(2*m), ytrue2(2*m);
  
  for (i=0; i < m; i++) {
    x(i) = i;
    x2(i) = i;
    x2(i+m) = i;
    dv(i) = i+1;
  }

  int descra[9];

  descra[0] = 3;  // triangular 
  descra[1] = 1;  // lower triangular
  descra[2] = 0;  // non-unit diagonal

for (int transp = 0; transp < 2; transp++ ) {
  // if transp == 0, operate with matrix, not transpose, else
  // if transp == 1, operate with matrix transpose
  int k = 2;                       // two right-hand-sides
  int unitd = 2;                   // row scaling by diag matrix (in dv())
  double alpha = atof(argv[1]);    // constant scaling
  double beta =  atof(argv[2]);    // constant scaling

// Calculate b so that exact solution is x2 = [0;1;2;3;4;0;1;2;3;4]
  if ( alpha == 0.0 ) {
     b = 0.0;
     ytrue = beta*x;
  }
  else {
     for (i=0;i<m;i++) b(i) = (1.0 - beta)*x(i)/alpha/dv(i);
     if (transp == 1 ) 
        b = Ac.trans_mult(b);
     else
        b = Ac*b;
     ytrue = x;
  }
  for (i=0;i<m;i++) {
     b2(i) = b(i);
     b2(i+m) = b(i);
     ytrue2(i) = ytrue(i);
     ytrue2(i+m) = ytrue(i);
  }

  y2 = x2;
  dcscsm (transp, m, k, unitd, &dv(0), alpha,
		   descra, cval, rowind, colptr,
		   &b2(0), m, beta, &y2(0), m, &work(0), m);


  if (verbose){ 
     cout << "CSCSM y2-ytrue2:  " << norm(y2-ytrue2) << endl;
     cout << "  transpose = " << transp << "  unitd = " << unitd  << endl;
  }

  if (norm(y2-ytrue2) > tolerance) {
    errcount++;
  }

  y2 = x2;
  dcsrsm (transp, m, k, unitd, &dv(0), alpha,
		   descra, rval, colind, rowptr,
		   &b2(0), m, beta, &y2(0), m, &work(0), m);

  if (verbose){ 
     cout << "CSRSM y2-ytrue2:  " << norm(y2-ytrue2) << endl;
     cout << "  transpose = " << transp << "  unitd = " << unitd  << endl;
  }

  if (norm(y2-ytrue2) > tolerance) {
    errcount++;
  }

  y2 = x2;
  dcoosm (transp, m, k, unitd, &dv(0), alpha,
                   descra, cval, indx, jndx, Ac.NumNonzeros(),
                   &b2(0), m, beta, &y2(0), m, &work(0), m);

  if (verbose){
     cout << "COOSM y2-ytrue2:  " << norm(y2-ytrue2) << endl;
     cout << "  transpose = " << transp << "  unitd = " << unitd  << endl;
  }

  if (norm(y2-ytrue2) > tolerance) {
    errcount++;
  }

  k = 2;                       // two right-hand-sides
  unitd = 3;                   // row scaling by diag matrix (in dv())
  alpha = atof(argv[1]);       // constant scaling
  beta =  atof(argv[2]);       // constant scaling

// Calculate b so that exact solution is x2 = [0;1;2;3;4;0;1;2;3;4]
  if ( alpha == 0.0 ) {
     b = 0.0;
     ytrue = beta*x;
  }
  else {
     for (i=0;i<m;i++) b(i) = (1.0 - beta)*x(i)/alpha;
     if (transp == 1 ) 
        b = Ac.trans_mult(b);
     else
        b = Ac*b;
     for (i=0;i<m;i++) b(i) = b(i)/dv(i);
     ytrue = x;
  }
  for (i=0;i<m;i++) {
     b2(i) = b(i);
     b2(i+m) = b(i);
     ytrue2(i) = ytrue(i);
     ytrue2(i+m) = ytrue(i);
  }

  y2 = x2;
  dcscsm (transp, m, k, unitd, &dv(0), alpha,
                   descra, cval, rowind, colptr,
                   &b2(0), m, beta, &y2(0), m, &work(0), m);

  if (verbose){ 
     cout << "CSCSM y2-ytrue2:  " << norm(y2-ytrue2) << endl;
     cout << "  transpose = " << transp << "  unitd = " << unitd  << endl;
  }

  if (norm(y2-ytrue2) > tolerance) {
    errcount++;
  }

  y2 = x2;
  dcsrsm (transp, m, k, unitd, &dv(0), alpha,
                   descra, rval, colind, rowptr,
                   &b2(0), m, beta, &y2(0), m, &work(0), m);

  if (verbose){ 
     cout << "CSRSM y2-ytrue2:  " << norm(y2-ytrue2) << endl;
     cout << "  transpose = " << transp << "  unitd = " << unitd  << endl;
  }
      
  if (norm(y2-ytrue2) > tolerance) {
    errcount++;
  }

  y2 = x2;
  dcoosm (transp, m, k, unitd, &dv(0), alpha,
                   descra, cval, indx, jndx, Ac.NumNonzeros(),
                   &b2(0), m, beta, &y2(0), m, &work(0), m);

  if (verbose){
     cout << "COOSM y2-ytrue2:  " << norm(y2-ytrue2) << endl;
     cout << "  transpose = " << transp << "  unitd = " << unitd  << endl;
  }

  if (norm(y2-ytrue2) > tolerance) {
    errcount++;
  }

}
  if (errcount > 0) {
    cout << "   There were " << errcount << " errors recorded in " ;
    for (i=0;i<argc;i++) cout << argv[i] << " " ;
    if (verbose == 0) {
      cout << " (run with -v to locate)" << endl;      
    }
  } else {
    if (verbose == 1) {
    cout << "   ++++++++++++++++++++++++++++++++++++++++++++++++++++++++" << endl;
    cout << "   +      No errors detected in Sparse BLAS routines:     +" << endl;
    cout << "   +   CompCol_LowerDiagSolve,  CompCol_UpperDiagSolve    +" << endl;
    cout << "   +   CompRow_LowerDiagSolve,  CompRow_UpperDiagSolve    +" << endl;
    cout << "   +              (multiple right-hand-sides)             +" << endl;
    cout << "   ++++++++++++++++++++++++++++++++++++++++++++++++++++++++" << endl;
    cout << endl;
    }
  }

  return errcount;
}
